#define _POSIX_SOURCE	1

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>

#ifdef SYSTEM_MALLOC
#include <sys/types.h>
#include <malloc.h>
#else
#include "malloc.h"
#endif /* SYSTEM_MALLOC */

#include <signal.h>

#undef CONTENTS

#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))

struct _node {
    size_t size;
    caddr_t memory;
    struct _node *next;
    char fill;
};
typedef struct _node node;

node *list = NULL;
unsigned long length = 0;
size_t maxbytes;
size_t sofar;
unsigned long malloc_calls = 0;
unsigned long free_calls = 0;
double malloc_time = 0;
double free_time = 0;


static void tvdiff(tv1, tv2, tvd)
    struct timeval *tv1;
    struct timeval *tv2;
    struct timeval *tvd;
{
    tvd->tv_sec = tv2->tv_sec - tv1->tv_sec;
    tvd->tv_usec = tv2->tv_usec - tv1->tv_usec;

    if (tv2->tv_sec > tv1->tv_sec) {
	tvd->tv_sec--;
	tvd->tv_usec += 1e6;
    }
} /* tvdiff */


static caddr_t timed_malloc(size)
    size_t size;
{
    struct timeval tv1, tv2, tvd;
    caddr_t retval;

    gettimeofday(&tv1, NULL);
    retval = malloc(size);
    gettimeofday(&tv2, NULL);

    tvdiff(&tv1, &tv2, &tvd);
    malloc_time += tvd.tv_sec + tvd.tv_usec / 1e6;

    malloc_calls++;

    return retval;
} /* timed_malloc */


static void timed_free(ptr)
    caddr_t ptr;
{
    struct timeval tv1, tv2, tvd;

    gettimeofday(&tv1, NULL);
    free(ptr);
    gettimeofday(&tv2, NULL);

    tvdiff(&tv1, &tv2, &tvd);
    free_time += tvd.tv_sec + tvd.tv_usec / 1e6;

    free_calls++;
} /* timed_free */


static void insert(head, n)
    node **head;
    node *n;
{
    node *c, *p;
    unsigned long skip = (length) ? (random() % length) : 0;
    unsigned long pos = 0;

    c = p = *head;
    while (pos < skip) {
	p = c;
	c = c->next;
	pos++;
    }

    if (c == p) {
	n->next = *head;
	*head = n;
    } else {
	p->next = n;
	n->next = c;
    }

    length++;
} /* insert */


static void delete(head, n)
    node **head;
    node *n;
{
    node *c, *p;

    c = p = *head;
    while ((c != NULL) && (c != n)) {
	p = c;
	c = c->next;
    }

    if (c == NULL) {
	fprintf(stderr, "mtest: trying to delete a nonexistant node!\n");
	return;
    }

    if (c == p) {
	*head = c->next;
    } else {
	p->next = c->next;
    }

    length--;
} /* delete */


static void allocate() {
    size_t s;
    node *n;

    n = (node *)timed_malloc(sizeof(node));
    if (n == NULL) {
	fprintf(stderr, "mtest: malloc(%d) failed!\n", sizeof(node));
	exit(1);
    }
    sofar += sizeof(node);

    insert(&list, n);

    s = ((random() % (maxbytes - sofar)) + 1) / 16;
    s = s ? s : 1;
    n->memory = timed_malloc(s);
    if (n->memory == NULL) {
	fprintf(stderr, "mtest: malloc(%d) failed!\n", s);
	exit(1);
    }
    sofar += s;

    n->size = s;

#ifdef CONTENTS
    n->fill = (char)(random() % 256);
    (void)memset((char *)n->memory, n->fill, n->size);
#endif /* CONTENTS */

} /* allocate */


static void deallocate() {
    node *t;
    unsigned long skip = (length) ? (random() % length) : 0;
    unsigned long pos = 0;

    t = list;
    while (pos < skip) {
	t = t->next;
	pos++;
    }

    if (t == NULL) {
	return;
    }

#ifdef CONTENTS
    for (pos = 0; pos < t->size; pos++) {
	if (t->memory[pos] != t->fill) {
	    fprintf(stderr, "mtest: memory has been corrupted!\n");
	    break;
	}
    }
#endif /* CONTENTS */

    timed_free(t->memory);
    sofar -= t->size;

    delete(&list, t);

    timed_free((caddr_t)t);
    sofar -= sizeof(node);
} /* deallocate */


static void mtest() {
    while(1) {
	if (sofar == 0) {
	    allocate();
	} else if (sofar + sizeof(node) >= maxbytes) {
	    deallocate();
	} else {
	    switch (random() % 12) {
		case 0: case 1: case 2: case 3: case 4:
		    allocate();
		    break;
		case 5: case 6: case 7: case 8: case 9:
		    deallocate();
		    break;
		case 10:
		    while (sofar + sizeof(node) < maxbytes) {
			allocate();
		    }
		    break;
		case 11:
		    while (sofar > 0) {
			deallocate();
		    }
		    break;
		default:
		    fprintf(stderr, "mtest: something horrible has happened!\n");
		    exit(1);
		    break;
	    }
	}
    }
} /* mtest */


static void printstats() {
    fprintf(stderr,
	    "\nmtest: %ld malloc calls in %4.2f seconds = %4.2f calls/second",
	    malloc_calls, malloc_time, (malloc_calls / malloc_time));
    fprintf(stderr,
	    "\nmtest: %ld free calls in %4.2f seconds = %4.2f calls/second",
	    free_calls, free_time, (free_calls / free_time));
} /* printstats */


static void interrupted() {
    printstats();
    fprintf(stderr, "\nmtest: interrupted!\n");
    exit(1);
} /* interrupted */


int main(argc, argv)
    int argc;
    char *argv[];
{
    float megs;

    (void)signal(SIGINT, interrupted);

    srandom((int)time(NULL));

    if (argc < 2) {
	fprintf(stderr, "mtest: usage: mtest #megabytes\n");
	exit(1);
    }

    sscanf(argv[1], "%f", &megs);

    if (megs <= 0.0) {
	fprintf(stderr, "mtest: positive number of megabytes, please\n");
	exit(1);
    }

    maxbytes = (size_t)(megs * 1024 * 1024);
    sofar = 0;

    mtest();

    return(0);
} /* main */
